// code by jph

class format {
public:
  static tensor<char> txt(char* name); // reads txt file
  static void bmp(char* name,const tensor<uchar>& img); // writes img in 24/32-bit/pixel bmp-file
  static tensor<uchar> bmp(char* name); // reads img in 24/32-bit/pixel bmp-file
};

tensor<char> format::txt(char* name) {
  FILE *fid=fopen(name,"rt");
  tensor<char> ret;
  if (fid) {
    fseek(fid,0,SEEK_END);
    ret=tensor<char>(1,ftell(fid));
    rewind(fid);
    fread(ret.data,sizeof(char),ret.numel,fid);
    fclose(fid);
  }
#ifdef NDEBUG
  else
#endif
  printf("%s -> %s\n",name,ret.toInfo());
  return ret;
}

// writes RGB and RGBA
void format::bmp(char* name,const tensor<uchar>& img) {
  assert(img.data);
  assert(img.size[0]==3||img.size[0]==4);
  FILE *fid=fopen(name,"wb");
  int num,sze,ofs=0x36;
  int mod=(img.size[0]*img.size[1])%4; // bgr. // TODO w/o parentheses
  if (fid) {
    num=0x4d42; fwrite(&num,2,1,fid); // identifier
    sze=ofs+img.size[2]*(img.size[0]*img.size[1]+(4-mod)*!!mod); // total file size
                fwrite(&sze,4,1,fid); // size of file (!)
    num=0;      fwrite(&num,4,1,fid); // unused
                fwrite(&ofs,4,1,fid); // pixel offset
    num=0x28;   fwrite(&num,4,1,fid); // size of header
                fwrite(&img.size[1],4,1,fid); // width
                fwrite(&img.size[2],4,1,fid); // height
    num=1;      fwrite(&num,2,1,fid); // color planes
    num=8*img.size[0];
                fwrite(&num,2,1,fid); // bits per pixel, 3bytes
    num=0;      fwrite(&num,4,1,fid); // no compression
                fwrite(&img.numel,4,1,fid); // size of data
    num=0xc40e; fwrite(&num,4,1,fid); // pixel per meter resolution horizontal
    num=0xc40e; fwrite(&num,4,1,fid); // pixel per meter resolution vertical
    num=0;      fwrite(&num,4,1,fid); // colors in palette
                fwrite(&num,4,1,fid); // colors important
    num=0;
    for (int c0=0;c0<img.size[2];++c0) {
      for (int c1=0;c1<img.size[1];++c1) {
                fwrite(&img.data[num+2],1,1,fid); // b
                fwrite(&img.data[num+1],1,1,fid); // g
                fwrite(&img.data[num+0],1,1,fid); // r
        if (img.size[0]==4) // this format works with Gimp 2.6.7  bmp format 32 bit "A8R8G8B8"
                fwrite(&img.data[num+3],1,1,fid); // a
        num+=img.size[0];
      }
      if (mod)
                fwrite(&ofs,4-mod,1,fid); // can write anything here!
    }    
    fclose(fid);
    printf("%s <-\tbitmap[RGB%sx%dx%d] = %d B (resid=%i)\n",name,img.size[0]==3?"":"A",img.size[1],img.size[2],sze,mod);
  } else
    printf("%s <- (error)\n",name);  
}

// reads RGB and RGBA
tensor<uchar> format::bmp(char* name) {
  tensor<uchar> img; // ndims==numel==0
  FILE *fid=fopen(name,"rb");
  uchar inp[4];
  int size[3];
  int num,sze,ofs=0x36,mod;  
  if (fid) {
    fread(inp,2,1,fid);   // identifier 0x4d42
    assert(inp[0]=='B'&&inp[1]=='M');
    fread(&sze,4,1,fid);  // size of file (!)
    fread(&num,4,1,fid);  // unused
    fread(&ofs,4,1,fid);  // pixel offset
    fread(&num,4,1,fid);  // size of header
    fread(&size[1],4,2,fid);  // width x height
    fread(&num,2,1,fid); // color planes    
    fread(&num,2,1,fid); // bits per pixel, 3bytes
    size[0]=(num&0xffff)>>3;
    assert(size[0]==3||size[0]==4);
    img=tensor<uchar>(3,size);
    mod=img.size[0]*img.size[1]%4; // bgr.    
    fread(&num,4,1,fid); // no compression
    assert(!num);
    fread(&num,4,1,fid); // size of data
    fread(&num,4,1,fid); // pixel per meter resolution horizontal
    fread(&num,4,1,fid); // pixel per meter resolution vertical
    fread(&num,4,1,fid); // colors in palette
    assert(!num);
    fread(&num,4,1,fid); // colors important
    num=0;
    for (int c0=0;c0<img.size[2];++c0) {
      for (int c1=0;c1<img.size[1];++c1) {
        fread(&img.data[num+2],1,1,fid); // b
        fread(&img.data[num+1],1,1,fid); // g
        fread(&img.data[num+0],1,1,fid); // r
        if (img.size[0]==4) // this format works with Gimp 2.6.7  bmp format 32 bit "A8R8G8B8"
          fread(&img.data[num+3],1,1,fid); // a
        num+=img.size[0];
      }
      if (mod)
        fread(&ofs,4-mod,1,fid); // can read anything here
    }
    fclose(fid);
    printf("%s ->\tbitmap[RGB%sx%dx%d] = %d B (resid=%i)\n",name,img.size[0]==3?"":"A",img.size[1],img.size[2],sze,mod);
  } else
    printf("%s <- (error)\n",name);  
  return img;
}
