// code by jph

#ifndef SET_INCLUDED
#define SET_INCLUDED

#define SET_RESIZE                  while (size%2==0||size%3==0||size%5==0) \
                                      ++size;

#define SET_EMPTY                   SET_RESIZE \
                                    table=tensor<linked<type>*>(1,&size); \
                                    for (int c1=0;c1<table.numel;c1++) \
                                      table(c1)=new linked<type>(); \
                                    numel=0;

#define SET_VARIANCE                tensor<double> occ(1,table.numel); \
                                    for (int c0=0;c0<table.numel;c0++) \
                                      occ.data[c0]=table(c0)->numel; \
                                    return occ.variance().data[0];

static char SET_BUFFER[1024];

template<class type>
class set {
  public:
    set<type>(void);
    set<type>(int size);
    set<type>(int size,uint (*func)(type));
    ~set<type>(void);
    int    has(const type& key);
    int    add(const type& key);
    int toggle(const type& key);
    tensor<type> toTensor(void);
    double variance(void);
    char* toInfo(void);
    char* toString(void); 
//----------------------------------------------------------------------
    tensor<linked<type>*> table;
    int numel;
    static uint map(type key); // default hash functions
  private:
    uint (*func)(type);
};

template<class type>
set<type>::set(void) { }

template<class type>
set<type>::set(int size) {
  SET_EMPTY
  func=map;
}

template<class type>
set<type>::set(int size,uint (*func)(type)) {
  SET_EMPTY
  this->func=func;
}

template<class type>
set<type>::~set(void) {
  for (int c1=0;c1<table.numel;c1++)
    delete table(c1);
}

template<class type>
int set<type>::add(const type& key) {
  uint adr=func(key)%table.numel;
  type tok;
  while (table(adr)->token(tok))
    if (tok==key) {
      table(adr)->reset(); // reset tokenizer
      return 1;
    }
  table(adr)->add(key);
  ++numel;
  return 0;
}

template<class type>
int set<type>::toggle(const type& key) {
  uint adr=func(key)%table.numel;
  type tok;
  while (table(adr)->token(tok))
    if (tok==key) {
      table(adr)->remove();
      table(adr)->reset(); // reset tokenizer
      --numel;
      return 1;
    }
  table(adr)->add(key);
  ++numel;
  return 0;
}

template<class type>
int set<type>::has(const type& key) {
  uint adr=func(key)%table.numel;
  type tok;
  while (table(adr)->token(tok))
    if (key==tok) {
      table(adr)->reset(); // reset tokenizer
      return 1;
    }
  return 0;
}

template<class type> 
tensor<type> set<type>::toTensor(void) {
  type tok;
  tensor<type> ret(1,numel);
  int count=0;
  for (int c0=0;c0<table.numel;++c0)
    while (table(c0)->token(tok))
      ret(count++)=tok;
      //count++;
      //ret(count)=tok;
      //  table(adr)->reset(); // reset tokenizer
      //  return 1;
      //}
  //ret.numel=ret.size[0]=numel;
  return ret;
}

template<class type> 
double set<type>::variance(void) {
  SET_VARIANCE
}

template<class type>
char* set<type>::toInfo(void) {
  sprintf(SET_BUFFER,"set<%i>(%i/%i occ=%5.1f%% var=%5.1f)",sizeof(type),numel,table.numel,numel*100/(double)table.numel,variance());
  return SET_BUFFER;
}

template<class type>
char* set<type>::toString(void) {
  return toTensor().toString();
}

uint set<uint>::map(uint key) {
  return key;
}

uint set<double>::map(double key) {
  return key;
}

uint set<int>::map(int key) {
  return key;
}

uint set<void*>::map(void* key) {
  char str[9];
  sprintf(str,"%p",key);  
  return strtol(str, 0, 16);
}

static int SET_PRIMES[12]={ // one makes it equivalent to map<int>
  1,29,541,3571,
  7919,48611,104729,287117,
  611953,1299709,7368787,15485863};


uint set<tensor<int>>::map(tensor<int> key) {
  assert(key.numel<9);
  int ret=0;
  for (int c0=0;c0<key.numel;++c0)
    ret+=key.data[c0]*SET_PRIMES[c0];
  return ret;
}

uint set<tensor<double>>::map(tensor<double> key) {
  assert(key.numel<9);
  double ret=0;
  for (int c0=0;c0<key.numel;++c0)
    ret+=key.data[c0]*SET_PRIMES[c0];
  return ret;
}

uint set<tensor<char>>::map(tensor<char> key) {
  int ret=0;
  for (int c0=0;c0<key.numel;++c0)
    ret+=key.data[c0];
  return ret;
}

#endif
